import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from pathlib import Path
import os
import re

# import utils
import datums
import visuals

from datums import *
from visuals import *


import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import warnings

import utils
import datums

def testfun():
	print('haddock')

def plot_cv(data=None, labels=None, line=True, scatter=False, errs=False, current_column=1, potential_column=0, err_column=3, xunits='V', yunits=r'$mA/cm^2$', export_name=None, export_type='png', fig_kw={}, err_kw={}, **plot_kw):
	"""
	Plot cyclic voltammetry data

	Parameters
	___________
	data: DataFrame or dict (default=None)
		Either a DataFrame containing CV data or a dict with DataFrames of CV data as values. Designed to be used seamlessly with the output of data.cv_process
	labels: array-like (default=None)
		List of labels to be used in the legend. If unspecified, keys of data are used as labels.
	line: bool (default=True)
		Whether to draw a line connecting the individual data points
	scatter: bool (default=False)
		Whether to draw a marker at each data point
	errs: bool (default=False)
		Whether to include an error bar at each data point
	current_column: int or str (default=1)
		Index or label of the column containing current data. Used only if automatic column identification fails
	potential_column: int or str (default=0)
		Index or label of the column containing potential data. Used only if automatic column identification fails
	err_column: int or str (default=3)
		Index or label of the column containing data to draw error bars. Used only if automatic column identification fails. Ignored if errs=False
	xunits: str (default='V')
		Units of the x-axis
	yunits: str (default=r'$mA/cm^2$')
		Units of the y-axis
	export_name: str, path object, or file-like (default=None)
		If specified, the figure will be saved as an image. Can either be a complete file path to save the image in a specific directory or a file name to save the image in the current directory
	export_type: str (default='png')
		File type to save the image as. Only used if export_name is specified and export_name does not include the file type
	fig_kw: dict
		Dict with keywords passed to the plt.subplots function to create the figure
	err_kw: dict
		Dict with keywords passed to the plt.errorbars function used to draw errorbars
	**plot_kw:
		all remaining keyword arguments are passed to the plt.plot or plt.scatter function used to draw the graphs

	Returns
	________
	fig: Figure
		Figure object containing all plot elements
	ax: Axes
		Axes object containing the plotted data
	"""
	if not utils.check_dict(data):
		data = {'key':data}
	labels = check_labels(data, labels)
	fig, ax = plt.subplots(**fig_kw)
	for l, df in zip(labels, list(data.values())):
		df.columns = utils.check_labels(df)
		x = datums.find_col(df, 'potential', potential_column)
		y = datums.find_col(df, 'current', current_column)
		yerr = check_errs(errs, df, 'current_err', err_column)
		plotter(ax, x, y, yerr, l, line, scatter, errs, err_kw, **plot_kw)
	if len(data) > 1:
		ax.legend(loc='best')
	ax.set_xlabel(build_axlabel('Potential', xunits))
	ax.set_ylabel(build_axlabel('Current Density', yunits))
	if export_name:
		fig_saver(export_name, export_type)
	return fig, ax

def polcurve(data=None, labels=None, line=True, scatter=True, errs=False, current_column=0, potential_column=1, err_column=3, xunits=r'$mA/cm^2$', yunits='V', export_name=None, export_type='png', fig_kw={}, err_kw={}, **plot_kw):
	"""
	Plot polarization curves using chronopotentiometry data

	Parameters
	___________
	data: DataFrame or dict (default=None)
		Either a DataFrame containing CP data or a dict with DataFrames of CP data as values. Designed to be used seamlessly with the output of data.cp_process
	labels: array-like (default=None)
		List of labels to be used in the legend. If unspecified, keys of data are used as labels.
	line: bool (default=True)
		Whether to draw a line connecting the individual data points
	scatter: bool (default=False)
		Whether to draw a marker at each data point
	errs: bool (default=False)
		Whether to include an error bar at each data point
	current_column: int or str (default=0)
		Index or label of the column containing current data. Used only if automatic column identification fails
	potential_column: int or str (default=1)
		Index or label of the column containing potential data. Used only if automatic column identification fails
	err_column: int or str (default=3)
		Index or label of the column containing data to draw error bars. Used only if automatic column identification fails. Ignored if errs=False
	xunits: str (default=r'$mA/cm^2$')
		Units of the x-axis
	yunits: str (default='V')
		Units of the y-axis
	export_name: str, path object, or file-like (default=None)
		If specified, the figure will be saved as an image. Can either be a complete file path to save the image in a specific directory or a file name to save the image in the current directory
	export_type: str (default='png')
		File type to save the image as. Only used if export_name is specified and export_name does not include the file type
	fig_kw: dict
		Dict with keywords passed to the plt.subplots function to create the figure
	err_kw: dict
		Dict with keywords passed to the plt.errorbars function used to draw errorbars
	**plot_kw:
		all remaining keyword arguments are passed to the plt.plot or plt.scatter function used to draw the graphs

	Returns
	________
	fig: Figure
		Figure object containing all plot elements
	ax: Axes
		Axes object containing the plotted data
	"""
	if not utils.check_dict(data):
		data={'key':data}
	labels = check_labels(data, labels)
	fig, ax = plt.subplots(**fig_kw)
	for l, df in zip (labels, list(data.values())):
		df.columns = utils.check_labels(df)
		x = datums.find_col(df, 'current', current_column)
		y = datums.find_col(df, 'potential', potential_column)
		yerr = check_errs(errs, df, 'potential_err', err_column)
		plotter(ax, x, y, yerr, l, line, scatter, errs, err_kw, **plot_kw)
	if len(data) > 1:
		ax.legend(loc='best')
	ax.set_xlabel(build_axlabel('Current Density', xunits))
	ax.set_ylabel(build_axlabel('Potential', yunits))
	if export_name:
		fig_saver(export_name, export_type)
	return fig, ax

def plot_cp_raw(data=None, labels=None, line=False, scatter=True, errs=False, current_column=2, potential_column=1,time_column=0, err_column=(4,5), xunits='s', yunits=('mA', 'V'), export_name=None, export_type='png', fig_kw={}, err_kw={}, **plot_kw):
	"""
	Plot raw chronopotentiometry data

	Note: it is strongly reccomended to use this function with data from only a single test and the default values of line, scatter, and errs to avoid overplotting.

	Parameters
	___________
	data: DataFrame or dict (default=None)
		Either a DataFrame containing CP data or a dict with DataFrames of CP data as values. Designed to be used seamlessly with the output of data.cp_raw
	labels: array-like (default=None)
		List of labels to be used in the legend. If unspecified, keys of data are used as labels.
	line: bool (default=True)
		Whether to draw a line connecting the individual data points
	scatter: bool (default=False)
		Whether to draw a marker at each data point
	errs: bool (default=False)
		Whether to include an error bar at each data point
	current_column: int or str (default=2)
		Index or label of the column containing current data. Used only if automatic column identification fails
	potential_column: int or str (default=1)
		Index or label of the column containing potential data. Used only if automatic column identification fails
	time_column: int or str (default=0)
		Index or label of the column containing time data. Used only if automatic column identification fails
	err_column: tuple or list of int or str (default=(4,5))
		Indices or labels of the columns containing data to draw error bars. Used only if automatic column identification fails. Ignored if errs=False
	xunits: str (default=r'$mA/cm^2$')
		Units of the x-axis
	yunits: tuple or list of str (default=('ma','V'))
		Units of the y-axis
	export_name: str, path object, or file-like (default=None)
		If specified, the figure will be saved as an image. Can either be a complete file path to save the image in a specific directory or a file name to save the image in the current directory
	export_type: str (default='png')
		File type to save the image as. Only used if export_name is specified and export_name does not include the file type
	fig_kw: dict
		Dict with keywords passed to the plt.subplots function to create the figure
	err_kw: dict
		Dict with keywords passed to the plt.errorbars function used to draw errorbars
	**plot_kw:
		all remaining keyword arguments are passed to the plt.plot or plt.scatter function used to draw the graphs

	Returns
	________
	fig: Figure
		Figure object containing all plot elements
	ax: Tuple
		Tuple of axes objects containing the plotted data
	"""
	if not utils.check_dict(data):
		data = {'key': data}
	labels = check_labels(data, labels)
	fig, ax1 = plt.subplots(**fig_kw)
	ax2 = ax1.twinx()
	color1 = 'tab:red'
	color2 = 'tab:blue'
	for l, df in zip (labels, list(data.values())):
		x = datums.find_col(df, 'time', time_column)
		y1 = datums.find_col(df, 'potential', potential_column)
		y2 = datums.find_col(df, 'current', current_column)
		yerr1 = check_errs(errs, df, 'current_err', err_column[0])
		yerr2 = check_errs(errs, df, 'potential_err', err_column[1])
		plotter(ax1, x, y1, yerr1, l, line, scatter, errs, err_kw, s=1, c=color1, **plot_kw)
		plotter(ax2, x, y2, yerr2, l, line, scatter, errs, err_kw, s=1, c=color2, **plot_kw)
	if len(data) > 1:
		ax1.legend(loc='best')
	# color = 'tab:red'
	ax1.set_xlabel(build_axlabel('Time', xunits))
	ax1.set_ylabel(build_axlabel('Current Density', yunits[0]))
	ax2.set_ylabel(build_axlabel('Potential', yunits[1]))
	ax1.tick_params(axis='y', labelcolor=color1)
	ax2.tick_params(axis='y', labelcolor=color2)
	if export_name:
		fig_saver(export_name, export_type)
	return fig, (ax1, ax2)

def build_axlabel(base, units):
	"""
	Generate axis label

	Auxilliary function to generate an axis label from the specified name and units:
	'Base [units]'

	Parameters
	___________
	base: str
		Axis name
	units: str
		Axis units

	Returns
	________
	label:
		Complete axis label
	"""
	label =  base + ' [' + units + ']'
	return label

def check_errs(errs, df, err_name, err_col):
	"""
	Check if data set contains error data and return valid error data

	Auxilliary function to check if the given data set contains data that can be used to draw error bars. Returns the error values if possible, otherwise returns an array of zeros.

	Parameters
	___________
	errs: bool
		Whether data set should be checked for error data. If False, an array of zeros is returned.
	df: DataFrame
		DataFrame that should contain error data
	err_name: str
		String of the form '[datatype]_std' (ex: 'current_std'). Used to parse column labels of df
	err_col: int or str
		Index or label of the column containing data to draw error bars. Used only if automatic column identification fails. Ignored if errs=False

	Returns
	________
	err: numpy array
		Array of values which can be used to draw error bars. If valid error values could not be found, an array of zeros of the same length as df is returned
	"""
	count = df.shape[0]
	if errs:
		try:
			err = datums.find_col(df, err_name, err_col)
		except:
			err = np.zeros(count)
			warnings.warn('Unable to use the specified error values', UserWarning)
	else:
		err = np.zeros(count)
	return err

def check_labels(data, labels=None):
	"""
	Check if the specified labels are valid

	Auxilliary functin to checks if the user-specified labels can be used to create the legend. If unspecified or lengths of data and labels are mismatched, dict keys of data are used as default labels.

	Parameters
	___________
	data: dict
		Dict with data to be plotted as values
	labels:list-like
		List of labels to be used in the legend. If unspecified, keys of data are used as labels.

	Returns
	________
	labels:
		Valid list of labels for use in the legend
	"""
	if labels and (len(labels) != len(data)):
		labels = list(data.keys())
		warnings.warn('labels and data must have the same length. Using default labels instead of specified labels', UserWarning)
	elif labels is None:
		labels = list(data.keys())
	return labels

def fig_saver(export_name, export_type='png'):
	"""
	Save the current figure

	Auxilliary function to save the current figure as an image.

	export_name: str, path object, or file-like
		File name to save the image as. Can either be a complete file path to save the image in a specific directory or a file name to save the image in the current directory
	export_type: str (default='png')
		File type to save the image as. Only used if export_name does not include the file type
	"""
	if '.' not in export_name:
		export_type = export_type.replace('.','')
		export_name = export_name + '.' + export_type
	plt.savefig(export_name, bbox_inches='tight')

def plotter(ax, x, y, e, l, line, scatter, errs, err_kw, **plot_kw):
	"""
	Plot data

	Auxilliary function to plot the specified data

	Parameters
	___________
	ax: Axes
		Axes object on which to plot the data
	x: array-like
		x values
	y: array-like
		y values
	e: array-like
		Error values used to draw error bars
	l: array-like
		Labels to be used in the legend
	line: bool
		Whether to draw a line connecting the individual data points
	scatter: bool
		Whether to draw a marker at each data point
	errs: bool
		Whether to include an error bar at each data point
	err_kw: dict
		Dict with keywords passed to the plt.errorbars function used to draw errorbars
	**plot_kw:
		all remaining keyword arguments are passed to the plt.plot or plt.scatter function used to draw the graphs
	"""
	if 'marker' not in plot_kw:
			plot_kw['marker'] = '.'
	if errs:
		if 'elinewidth' not in err_kw:
			err_kw['elinewidth'] = 0.5
		if 'capthick' not in err_kw:
			err_kw['capthick'] = 0.5
		if 'capsize' not in err_kw:
			err_kw['capsize'] = 3
		if line and scatter:
			ax.errorbar(x, y, e, label=l, **err_kw, **plot_kw)
		elif line:
			plot_kw.pop('marker')
			ax.errorbar(x, y, e, label=l, **err_kw, **plot_kw)
		else:
			plot_kw['ls'] = ''
			ax.errorbar(x, y, e, label=l, **err_kw, **plot_kw)
	else :
		if line and scatter:
			ax.plot(x, y, label=l, **plot_kw)
		elif line:
			plot_kw.pop('marker')
			ax.plot(x, y, label=l, **plot_kw)
		else:
			ax.scatter(x, y, label=l, **plot_kw)


import numpy as np
import pandas as pd
import os
import re
import warnings

valid_types = ['csv', 'xls', 'xlsx', 'txt']
export_types = ['csv', 'xls', 'xlsx']
excel_types = ['xls', 'xlsx']
csv_types = ['csv', 'txt']
dlm_default = '\t'
default_savetype = 'csv'

label_dict = {'v':'v', 'ma':'i', 'a':'i', 's':'t', 'mv':'v', 'v vs. sce':'v', 'mv vs. sce':'v',
				'v vs. she':'v', 'mv vs. she':'v'}
default_names = ['tintin', 'snowy', 'haddock', 'calculus', 'castafiore', 'thomson', 'thompson']

def check_type(filetype):
	"""
	Check if the filetype is currently supported for import

	fuelcell currently supports import of csv, xls, xlsx, and txt files

	Parameters
	___________
	filetype: str
		Filetype to be checked

	Returns
	________
	filetype: str
		If the given filetype string is valid, it is returned in lowercase and stripped of '.' characters 

	"""
	filetype = filetype.lower().replace('.', '')
	if filetype not in valid_types:
		raise ValueError('Supported filetypes ' + ', '.join(valid_types))
	return filetype

def check_export_type(filetype):
	"""
	Check if the filetype is currently supported for export

	fuelcell currently supports export of csv, xls, and xlsx files

	Parameters
	___________
	filetype: str
		Filetype to be checked

	Returns
	________
	filetype: str
		If the given filetype string is valid, it is returned in lowercase and stripped of '.' characters. If the filetype is not valid, returns 'csv'

	"""
	filetype = filetype.lower().replace('.','')
	if filetype not in export_types:
		warnings.warn(filetype + 'is not a supported export format. Currently supported for export: ' + ', '.join(export_types), UserWarning)
		filetype = 'csv'
	return filetype

def check_list(var):
	"""
	Checks if var is a python list or numpy array

	Parameters
	___________
	var:
		variable to be checked

	Returns
	________
	result: bool
		True if var is a list or numpy array, False otherwise

	"""
	result = (type(var) == list) or (type(var) == np.ndarray)
	return result

def check_dict(var):
	"""
	Checks if var is a python dictionary

	Parameters
	___________
	var:
		variable to be checked

	Returns
	________
	result: bool
		True if var is a dict, False otherwise

	"""	
	result = type(var) == dict
	return result

def check_str(var):
	"""
	Checks if var is a string

	Parameters
	___________
	var:
		variable to be checked

	Returns
	________
	result: bool
		True if var is a string, False otherwise

	"""	
	result = type(var) == str
	return result

def check_float(var):
	"""
	Checks if var is a float

	Parameters
	___________
	var:
		variable to be checked

	Returns
	________
	result: bool
		True if var is a float, False otherwise

	"""	
	result = type(var) == float
	return result

def check_int(var):
	"""
	Checks if var is an integer

	Parameters
	___________
	var:
		variable to be checked

	Returns
	________
	result: bool
		True if var is an int, False otherwise

	"""	
	result = type(var) == int
	return result

def check_scalar(var):
	"""
	Checks if var is a scalar

	If the variable has a valid length (ie if len(var) is a valid call), it is considered not to be a scalar.

	Parameters
	___________
	var:
		variable to be checked.

	Returns
	________
	result: bool
		True if var is a scalar, False otherwise

	"""	
	try:
		len(var)
		return False
	except:
		return True

def check_savedir(folder):
	"""
	Check if the specified folder exists

	Checks if the specified folder exists. If the folder does not exist and cannot be created, returns a path to a folder named 'processed' within the current directory

	Parameters
	___________
	folder: str, path object, or path-like
		Folder to be checked

	Returns
	________
	path: str
		Path to folder. Defaults to 'processed' if any errors occur.
	"""
	if os.path.exists(folder):
		path = os.path.realpath(folder)
	else:
		try:
			os.mkdir(folder)
			path = os.path.realpath(folder)
		except FileNotFoundError:
			warnings.warn('Unable to save to ' + folder + '. Saving data to the current directory', UserWarning)
			if os.path.exists('processed'):
				path = os.path.realpath('processed')
			else:
				os.mkdir('processed')
				path = os.path.realpath('processed')
	return path

def check_labels(data):
	"""
	Parses column labels to determine the data stored in each column

	Current, voltage, time, etc. columns are determined from the units based on standard EC Lab conventions. If a column can be identified, its name is converted to the standard heading ('i' for current, 'v' for potential, etc.), otherwise, the namne is unchanged.

	Parameters
	___________
	data: DataFarame
		DataFrame in need of column identification

	Returns
	________
	newcols: list
		List containing the new column labels
	"""	
	cols = [c.lower() for c in data.columns]
	newcols = [];
	for c in cols:
		try:
			units = c.split('/')[1]
			if units in label_dict.keys():
				newcols.append(label_dict[units])
			else:
				newcols.append(c)
		except:
			newcols.append(c)
	return newcols

def get_files(path=None, pattern='', filetype=''):
	"""
	Find all files matching the desired criteria

	Determines which files in the specified directory match the given regular expression and/or are of the given filetype

	Parameters
	___________
	path: str, path object, or path-like (default=None)
		Directory in which files will be searched for. If unspecified, the current directory will be used.
	patter: str (default='')
		Regular expression. Only files matching this regular will be found. If unspecified, this parameter is ignored.
	filetype: str or regex (default='')
		Only files of this filetype will be found. If unspecified, filetype is ignored when finding files
	Returns
	________
	files: list
		List of all files in the specified directory matching the given criteria

	"""	
	files = os.listdir(path)
	files.sort()
	if pattern:
		files = [f for f in files if re.match(pattern, f)]
	if filetype:
		filetype = filetype.lower().replace('.', '')
		files = [f for f in files if re.match(r'.*\.'+filetype, f)]
	return files

def read_file(filename, dlm=dlm_default):
	"""
	Loads the specified file as a pandas DataFrame

	The specified file must be of one of the types supported for import. fuelcell currently supports csv, xls, xlsx, and txt files. If the file has a valid filetype but cannot be imported for some reason, a warning is raised.

	Parameters
	___________
	filename: str, path object, or path-like
		Name of a file in the current directory, or a complete path to the desired file.
	dlm: str (default='\t')
		Delimiting character if the file is a text file. Defaults to '\t' (tab-delimiting).

	Returns
	________
	name: str
		Filename with the filetype removed.
	data: DataFrame
		Pandas DataFrame containing the data read from the file.
	"""	
	data = None
	name = None
	try:
		name = os.path.basename(filename)
		name, filetype = name.split('.')
		filetype = check_type(filetype)
		if filetype in excel_types:
			data = pd.read_excel(filename)
		elif filetype in csv_types:
			if filetype == 'csv':
				data = pd.read_csv(filename)
			elif filetype == 'txt':
				data = pd.read_csv(filename, delimiter=dlm)
		return name, data
	except:
		if not os.path.isdir(filename):
			if filename.split('.')[0] in valid_types:
				warnings.warn(f'Unable to read {os.path.basename(filename)}', UserWarning)
	return name, data

def get_testdir():
	fcdir = os.path.dirname(os.path.realpath(__file__))
	datapath = os.path.join(fcdir, 'testdata')
	return datapath

def save_data(data, filename=None, folder=None):
	"""
	Save data to a local file

	Saves data stored in a pandas DataFrame as an excel or csv file. Valid filetypes to save as are xls, xlsx, and csv (default is csv). If the specified filename already exists, the name is modified to avoid unintentionally overwriting data. If the specified folder does not exist and cannot be created, the file is saved to a folder named 'processed' within the current director.

	Parameters
	___________
	data: DataFrame
		Data to be saved
	filename: str, path object, or file-like (default=None)
		Filename to save the data as. Either a full filename or complete path to an individual file. If the file already exists, the specified name is modified to avoid overwriting data. If unspecified, an arbitrary name is used.
	folder: str, path obbject, or path-like (default=None)
		Folder in which data will be saved. If unspecified and folder cannot be determined from filename, or the specified folder is invalid and cannot be created, defaults to 'processed'
	"""
	if filename:
		path, name = os.path.split(filename)
		name, fmt = name.split('.')
		filetype = check_export_type(fmt)
		name = name + '.' + filetype
	else:
		name = np.random.choice(default_names) + str(np.random.randint(100))
		if folder:
			path = folder
		else:
			path = 'processed'
		filetype = default_savetype
		warnings.warn('Filename unspecified. Saving data as ' + name + '.' + filetype)
	if path:
		savedir = check_savedir(path)
	elif folder:
		savedir = check_savedir(folder)
	else:
		if os.path.exists('processed'):
			savedir = os.path.realpath('processed')
		else:
			os.mkdir('processed')
			savedir = os.path.realpath('processed')
	full_path = os.path.join(savedir, name)
	if os.path.exists(full_path):
		while os.path.exists(full_path):
			name = name.replace('.', str(np.random.randint(1000))+'.')
			full_path = os.path.join(savedir, name)
		warnings.warn('Saving data as ' + name + ' to avoid overwriting existing file')
	if filetype in excel_types:
		data.to_excel(full_path, index=False)
	else:
		data.to_csv(full_path, index=False)


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
  name = 'fuelcell',
  version = '1.3.0',
  description = 'data processing for fuel cell and electrolyzer experiments',
  long_description = long_description,
  long_description_content_type = 'text/markdown',
  url = 'https://github.com/samaygarg/fuelcell',
  author = 'Samay Garg',
  author_email = 'samay.garg@berkeley.edu',
  license = 'MIT',
  classifiers=[
    'Development Status :: 4 - Beta',
    'Programming Language :: Python :: 3'
  ],
  keywords = ['tintin', 'snowy', 'haddock', 'fuelcell'],
  packages=setuptools.find_packages(),
  python_requires='>=3'
)


import numpy as np
import pandas as pd
import os
import re

import utils

dlm_default = utils.dlm_default
col_default_labels = {'current':'i', 'potential':'v', 'time':'t', 'current_err':'i_sd', 'potential_err':'v_sd'}
col_default_ids = {'current':1, 'potential':2, 'time':0, 'current_err':2, 'potential_err':3}
ref_electrodes = {'she':0, 'sce':0.241}


def testfun():
	print('thompson')

def ca_raw(filename=None, folder=None, pattern='', filetype='', delimiter=dlm_default):
	"""
	Loads chronoamperometry data

	Efficient way to load multiple CA files at once;equivalent to calling load_data and specifying expt_type='ca'. If called with no arguments, loads all ca files in the present folder.

	Parameters
	___________
	filename: str, path object, or file-like (default=None)
		Full filename of a file in the present directory or a complete path to an individual file. If filename is specified, all other arguments except delimiter are ignored.
	folder: str, path object, or path-like (default=None)
		Directory in which data files are stored. If none, defaults to the present directory.
	pattern: str or regex
		If specified, only files matching this pattern in the specified folder are loaded. Ignored if filename is specified.
	filetype : str
		Any supported filetype. Only files of the specified file type will be loaded. Can be used in conjunction with pattern or expt_type.
	delimiter : char (default = '\t')
		Delimiting character if the file is a text file. Defaults to '\t' (tab-delimiting).

	Returns
	________
	data: DataFrame or dict
		If a single file is read, it is returned directly as a pandas DataFrame. If multiple files are read, a dictionary is returned with the file names as keys and DataFrames as values.
	"""
	data = load_data(filename, folder, pattern, 'ca', filetype, delimiter)
	return data

def ca_process(data=None, current_column=1, potential_column=2, export_data=False, threshold=5, min_step_length=50, pts_to_average=300, pyramid=False, area=5, reference=None, **kwargs):
	"""
	Processes chronoamperometry data

	Can either process pre-loaded data or load and process data files. If called with no arguments, loads and processes all 'ca' files in the present folder. See process_steps for details on the operations performed.

	Parameters
	___________
	data: DataFrame or dict
		Either a DataFrame containing CA data or a dict with DataFrames of CA data as values. If unspecified, data will be loaded using ca_raw before processing. Designed to be used seamlessly with the output of ca_raw
	current_column : int or str (default=1)
		Index or label of the column containing current data. Used only if automatic column identification fails
	potential_column : int or str (default=2)
		Index or label of the column containing potential data. Used only if automatic column identification fails
	threshold: int (default=5)
		Minimum consecutive absolute difference which constitutes a step
	min_step_length: int (default=25)
		Minimum length of the arrays which result from spliting the intial array. Arrays shorter than this value will be discarded
	pts_to_average: int (default=300)
		Steady-state average and sd are calculated using the last pts_to_average values of the array. Default is 300 points, which is the last 30 seconds of each hold at the instrument's default collection rate of 10 Hz.
	pyramid: bool (default=True)
		Specifies whether the current is ramped in both directions. Set pyramid=False if only ramping up or only ramping down.
	area: int or float (default=5)
		Geometric active area of the MEA. Scaling factor to convert current to current density.
	reference: {'she', 'sce'}, int, or float (default='she')
		Either a string identifying the reference electrode (ie 'she' or 'sce'), or the potential of the reference electrode used. sce=0.241
	**kwargs:
		Remaining arguments are passed to ca_raw to load data
	"""
	if data is None:
		data = ca_raw(**kwargs)
	if not utils.check_dict(data):
		data = {'key': data}
	newdata = dict()
	for k, df in data.items():
		ca_processed = process_steps(df, potential_column, current_column, threshold, min_step_length, pts_to_average, pyramid, 'ca', area, reference)
		newdata[k] = ca_processed
		if export_data:
			utils.save_data(ca_processed, k+'.csv', 'processed')
	if len(newdata) == 1:
		newdata = list(newdata.values())[0]
	return newdata

def cp_raw(filename=None, folder=None, pattern='', filetype='', delimiter=dlm_default):
	"""
	Loads chronopotentiometry data

	Efficient way to load multiple CP files at once;equivalent to calling load_data and specifying expt_type='cp'. If called with no arguments, loads all cp files in the present folder.

	Parameters
	___________
	filename: str, path object, or file-like (default=None)
		Full filename of a file in the present directory or a complete path to an individual file. If filename is specified, all other arguments except delimiter are ignored.
	folder: str, path object, or path-like (default=None)
		Directory in which data files are stored. If none, defaults to the present directory.
	pattern: str or regex
		If specified, only files matching this pattern in the specified folder are loaded. Ignored if filename is specified.
	filetype : str
		Any supported filetype. Only files of the specified file type will be loaded. Can be used in conjunction with pattern or expt_type.
	delimiter : char (default = '\t')
		Delimiting character if the file is a text file. Defaults to '\t' (tab-delimiting).

	Returns
	________
	data: DataFrame or dict
		If a single file is read, it is returned directly as a pandas DataFrame. If multiple files are read, a dictionary is returned with the file names as keys and DataFrames as values.
	"""
	data = load_data(filename, folder, pattern, 'cp', filetype, delimiter)
	return data

def cp_process(data=None, current_column=1, potential_column=2, export_data=False, threshold=5, min_step_length=25, pts_to_average=300, pyramid=True, area=5, reference=None, **kwargs):
	"""
	Processes chronopotentiometry data

	Can either process pre-loaded data or load and process data files. If called with no arguments, loads and processes all 'cp' files in the present folder. See process_steps for details on the operations performed.

	Parameters
	___________
	data: DataFrame or dict
		Either a DataFrame containing CP data or a dict with DataFrames of CP data as values. If unspecified, data will be loaded using cp_raw before processing. Designed to be used seamlessly with the output of cv_raw
	current_column : int or str (default=1)
		Index or label of the column containing current data. Used only if automatic column identification fails
	potential_column : int or str (default=2)
		Index or label of the column containing potential data. Used only if automatic column identification fails
	threshold: int (default=5)
		Minimum consecutive absolute difference which constitutes a step
	min_step_length: int (default=25)
		Minimum length of the arrays which result from spliting the intial array. Arrays shorter than this value will be discarded
	pts_to_average: int (default=300)
		Steady-state average and sd are calculated using the last pts_to_average values of the array. Default is 300 points, which is the last 30 seconds of each hold at the instrument's default collection rate of 10 Hz.
	pyramid: bool (default=True)
		Specifies whether the current is ramped in both directions. Set pyramid=False if only ramping up or only ramping down.
	area: int or float (default=5)
		Geometric active area of the MEA. Scaling factor to convert current to current density.
	reference: {'she', 'sce'}, int, or float (default='she')
		Either a string identifying the reference electrode (ie 'she' or 'sce'), or the potential of the reference electrode used. sce=0.241
	**kwargs:
		Remaining arguments are passed to cp_raw to load data
	"""
	if data is None:
		data = cp_raw(**kwargs)
	if not utils.check_dict(data):
		data = {'key': data}
	newdata = dict()
	for k, df in data.items():
		cp_processed = process_steps(df, current_column, potential_column, threshold, min_step_length, pts_to_average, pyramid, 'cp', area, reference)
		newdata[k] = cp_processed
		if export_data:
			utils.save_data(cp_processed, k+'.csv', 'processed')
	if len(newdata) == 1:
		newdata = list(newdata.values())[0]
	return newdata

def cv_raw(filename=None, folder=None, pattern='', filetype='', delimiter=dlm_default):
	"""
	Loads cyclic voltammetry data

	Efficient way to load multiple CV files at once;equivalent to calling load_data and specifying expt_type='cv'. If called with no arguments, loads all cv files in the present folder.

	Parameters
	___________
	filename: str, path object, or file-like (default=None)
		Full filename of a file in the present directory or a complete path to an individual file. If filename is specified, all other arguments except delimiter are ignored.
	folder: str, path object, or path-like (default=None)
		Directory in which data files are stored. If none, defaults to the present directory.
	pattern: str or regex
		If specified, only files matching this pattern in the specified folder are loaded. Ignored if filename is specified.
	filetype : str
		Any supported filetype. Only files of the specified file type will be loaded. Can be used in conjunction with pattern or expt_type.
	delimiter : char (default = '\t')
		Delimiting character if the file is a text file. Defaults to '\t' (tab-delimiting).

	Returns
	________
	data: DataFrame or dict
		If a single file is read, it is returned directly as a pandas DataFrame. If multiple files are read, a dictionary is returned with the file names as keys and DataFrames as values.
	"""
	data = load_data(filename, folder, pattern, 'cv', filetype, delimiter)
	return data

def cv_process(data=None, current_column=1, potential_column=0, export_data=False, area=5,  **kwargs):
	"""
	Processes cyclic voltammetry data

	Can either process pre-loaded data or load and process data files. If called with no arguments, loads and processes all 'cv' files in the present folder. Peforms the following operations in order:
	1. Parse column labels to find columns containing current and potential data. If parsing fails, specified labels/indices are used
	2. Convert current to current density using the specified area

	Parameters
	__________
	data : DataFrame or dict
		Either a DataFrame containing CV data or a dict with DataFrames of CV data as values. If unspecified, data will be loaded using cv_raw before processing. Designed to be used seamlessly with the output of cv_raw
	area : int or float (default=5)
		Geometric active area of the MEA. Scaling factor to convert current to durrent density
	current_column : int or str (default=1)
		Index or label of the column containing current data. Used only if automatic column identification fails
	potential_column : int or str (default=0)
		Index or label of the column containing potential data. Used only if automatic column identification fails
	**kwargs:
		Remaining arguments are passed to cv_raw to load data
	
	Returns
	_______
	newdata: DataFrame or dict
		If a single file is processed, it is returned directly as a pandas DataFrame. If multiple files are orocessed, a dictionary is returned with the file names as keys and DataFrames as values.
	"""
	if data is None:
		data = cv_raw(**kwargs)
	if not utils.check_dict(data):
		data = {'key': data}
	newdata = dict()
	for k, df in data.items():
		current = find_col(df, 'current', current_column)
		current = current / area
		potential = find_col(df, 'potential', potential_column)
		cv_processed = pd.DataFrame({'i':current,'v':potential})
		newdata[k] = cv_processed
		if export_data:
			utils.save_data(cv_processed, k+'.csv', 'processed')
	if len(newdata) == 1:
		newdata = list(newdata.values())[0]
	return newdata

def array_apply(arr, func, **kwargs):
	"""
	Applies a function to each value in an array

	Parameters
	___________
	arr: array-like
		Array to which function will be applied
	func: function
		function to apply to each value
	**kwargs:
		Any other arguments required by the specified function

	Returns
	________
	result: numpy array
		Array of the transformed values
	"""
	result =  np.array([func(a, **kwargs) for a in arr])
	return result

def avg_last_pts(arr, numpts=300):
	"""
	Average of the last several values of an array

	Auxilliary function to compute the average of an array accross the last several data points. Useful for obtaining a steady-state average.

	Parameters
	___________
	arr: list or numpy array
		Array of values used to compute the average.
	numpts: int (default=300)
		Average is calculated using the last numpts values of the array.

	Returns
	________
	avg: float
		Average of the last several values
	"""
	if type(arr) == list:
		arr = np.array(arr)
	avg = np.mean(arr[-numpts:])
	return avg

def find_col(data, col_type, label=None):
	"""
	Finds column containing the desired measurement

	Parameters
	___________
	data: DataFrame
		DataFrame containing the full data set
	col_type: {'current', 'potential', 'time', 'current_err', 'potential_err'}
		Type of data being searched for
	label: str or int (default=None)
		Label or index of the desired column if the column label cannot be automatically parsed

	Returns
	________
	col: numpy array
		Array of the desired measurement values
	"""
	default_label = col_default_labels[col_type]
	default_id = col_default_ids[col_type]
	newdf = data.copy()
	newdf.columns = utils.check_labels(newdf)
	if default_label in newdf.columns:
		col = newdf[default_label]
	elif label:
		if utils.check_str(label):
			col = newdf[label]
		else:
			col = newdf.iloc[:,label]
	else:
		col = newdf.iloc[:, default_id]
	col = np.array(col)
	return col

def find_steps(arr, threshold=5):
	"""
	Find indices at which an array of roughly stepwise data changes values

	Auxilliary function to find the points at which curren/voltage is stepped up or down during and experiment. Identifies 'steps' by determining the points at which the consecutive absolute difference between array values is greater than a specified threshold.

	Parameters
	___________
	arr: list or numpy array
		Array of roughly stepwise data
	threshold: int or float  (default=5)
		Minimum consecutive absolute difference which constitutes a step

	Returns
	________
	splits: numpy array
		Indices at which the array steps up or down
	"""
	if type(arr) == list:
		arr = np.array(arr)
	diffs = np.abs(np.diff(arr))
	splits = np.where(diffs > threshold)[0] + 1
	return splits

def load_data(filename=None, folder=None, pattern='', expt_type='', filetype='', delimiter=dlm_default):
	"""
	Loads data file(s) as a pandas DataFrame

	Function to load data files as a pandas DataFrame. If called with no
	arguments, loads all supported data files in the present folder.

	Parameters
	___________
	filename: str, path object, or file-like (default=None)
		Full filename of a file in the present directory or a complete path to an individual file. If filename is specified, all other arguments except delimiter are ignored.
	folder: str, path object, or path-like (default=None)
		Directory in which data files are stored. If none, defaults to the present directory.
	pattern: str or regex
		If specified, only files matching this pattern in the specified folder are loaded. Ignored if filename is specified.
	expt_type: str (default='')
		Alternative to specifying pattern; ignored if pattern is specified. All files containing expt_type anywhere in the file name will be loaded. Ex: to load all chronopotentiometry files, specify expt_type='cp'.
	filetype : str
		Any supported filetype. Only files of the specified file type will be loaded. Can be used in conjunction with pattern or expt_type.
	delimiter : char (default = '\t')
		Delimiting character if the file is a text file. Defaults to '\t' (tab-delimiting).

	Returns
	________
	data: DataFrame or dict
		If a single file is read, it is returned directly as a pandas DataFrame. If multiple files are read, a dictionary is returned with the file names as keys and DataFrames as values.
	"""
	data = None
	if filename:
		name, data = utils.read_file(filename, delimiter)
	else:
		if folder:
			dirpath = os.path.realpath(folder)
		else:
			dirpath = os.getcwd()
		if expt_type and not pattern:
			pattern = r'.*' + expt_type + r'.*'
		files = utils.get_files(dirpath, pattern, filetype)
		data = dict()
		for f in files:
			path = os.path.join(dirpath, f)
			name, df = utils.read_file(path, delimiter)
			if df is not None:
				data[name] = df
		if len(data) == 1:
			data = list(data.values())[0]
	return data

def process_steps(data, control_column=0, response_column=1, threshold=5, min_step_length=25, pts_to_average=300, pyramid=True, expt_type='cp', area=None, reference='she'):
	"""
	Processes stepwise data (ex chronopotentiometry and chronoamperometry data)

	Performs the following operations in order:
	1. Parse column labels to find the columns containing the desired data. If automatic parsing fails, the specified column labels/indices are used.
	2. Find the points at which the independent variable steps up or down using the specified threshold value.
	3. Split the data  at these split points. Holds with fewer than the minimum number of points are filtered out to account for outliers. Note: splitting is based upon values of the independent variable, so this will not result in the loss of meaningful data.
	4. Average and standard deviation are calculated using the last several points of each step to obtain steady-state values.
	5. If the indpendent variable is ramped up and down, the steady-state average and standard deviation are calculated across both the ramp-up and ramp-down holds. 
	6. Any necessary scaling or transformations are performed (ex. reference electrode correction, coversion to current density, etc.)

	Parameters
	___________
	data: DataFrame
		Dataframe containing the data to be processed
	control_column: int or str (default=0)
		Label or index of the column containing the control/independent variable values
	response_column: int or str (default=1)
		Label or index of the column cotaining the response/dependent variabble values
	threshold: int (default=5)
		Minimum consecutive absolute difference which constitutes a step
	min_step_length: int (default=25)
		Minimum length of the arrays which result from spliting the intial array. Arrays shorter than this value will be discarded
	pts_to_average: int (default=300)
		Steady-state average and sd are calculated using the last pts_to_average values of the array. Default is 300 points, which is the last 30 seconds of each hold at the instrument's default collection rate of 10 Hz.
	pyramid: bool (default=True)
		Specifies whether the current is ramped in both directions. Set pyramid=False if only ramping up or only ramping down.
	expt_type: {'cp', 'ca'} (default='cp')
		Specifies the type of experiment being analyzed. This is used to determine which variables are the control and response variables.
	area: int or float (default=5)
		Geometric active area of the MEA. Scaling factor to convert current to current density.
	reference: {'she', 'sce'}, int, or float (default='she')
		Either a string identifying the reference electrode (ie 'she' or 'sce'), or the potential of the reference electrode used. sce=0.241
	"""
	if expt_type == 'ca':
		control_var = 'potential'
		response_var = 'current'
	elif expt_type == 'cp': 
		control_var = 'current'
		response_var = 'potential'
	control = np.array(find_col(data, control_var, control_column))
	response = np.array(find_col(data, response_var, response_column))
	split_pts = find_steps(control, threshold=threshold)
	control_steps = split_and_filter(control, split_pts, min_length=min_step_length)
	response_steps = split_and_filter(response, split_pts, min_length=min_step_length)
	control_avg = array_apply(control_steps, avg_last_pts, numpts=pts_to_average)
	response_avg = array_apply(response_steps, avg_last_pts, numpts=pts_to_average)
	control_std = array_apply(control_steps, std_last_pts, numpts=pts_to_average)
	response_std = array_apply(response_steps, std_last_pts, numpts=pts_to_average)	
	# if expt_type == 'ca':
	# 	split_pts = find_steps(potential, threshold=threshold)
	# elif expt_type == 'cp':
	# 	split_pts = find_steps(current, threshold=threshold)
	# current_steps = split_and_filter(current, split_pts, min_length=min_step_length)
	# potential_steps = split_and_filter(potential, split_pts, min_length=min_step_length)
	# current_avg = array_apply(current_steps, avg_last_pts, numpts=pts_to_average)
	# potential_avg = array_apply(potential_steps, avg_last_pts, numpts=pts_to_average)
	# current_std = array_apply(current_steps, std_last_pts, numpts=pts_to_average)
	# potential_std = array_apply(potential_steps, std_last_pts, numpts=pts_to_average)
	if pyramid:
		sort_idx = np.argsort(control_avg)
		control_avg = control_avg[sort_idx]
		response_avg = response_avg[sort_idx]
		control_std = control_std[sort_idx]
		response_std = response_std[sort_idx]
		split_pts = find_steps(control_avg, threshold=2)
		# if expt_type == 'ca':
		# 	split_pts = find_steps(potential_avg, threshold=2)
		# elif expt_type == 'cp':
		# 	split_pts = find_steps(current_avg, threshold=2)
		control_steps = split_and_filter(control_avg, split_pts)
		response_steps = split_and_filter(response_avg, split_pts)
		control_std_steps = split_and_filter(control_std, split_pts)
		response_std_steps = split_and_filter(response_std, split_pts)
		control_avg = array_apply(control_steps, np.mean)
		response_avg = array_apply(response_steps, np.mean)
		control_std = array_apply(control_std_steps, std_agg)
		response_std = array_apply(response_std_steps, std_agg)
	# current_avg = current_avg / area
	# current_std = current_std / area
	if expt_type == 'ca':
		if reference:
			control_avg = electrode_correct(control_avg, reference)
		response_avg = response_avg / area
		response_std = response_std / np.sqrt(area)
		processed = pd.DataFrame({'i':response_avg, 'v':control_avg, 'i_sd':response_std, 'v_sd':control_std})
	elif expt_type == 'cp':
		if reference:
			response_avg = electrode_correct(response_avg, reference)
		control_avg = control_avg / area
		control_std = control_std / np.sqrt(area)
		processed = pd.DataFrame({'i':control_avg, 'v':response_avg, 'i_sd':control_std, 'v_sd':response_std})
	return processed

def electrode_correct(arr, ref='she'):
	"""
	Corrects for the reference electrode

	Parameters
	___________
	arr: list or numpy array
		Array of potential values to which the correction will be applied
	ref: {'she', 'sce'}, int, or float (default='she')
		Either a string identifying the reference electrode (ie 'she' or 'sce'), or the potential of the reference electrode used. sce=0.241
	"""
	if type(arr) == list:
		arr = np.array(arr)
	corrected = arr
	if utils.check_str(ref):
		ref = ref.lower()
		try:
			corrected += ref_electrodes[ref]
		except KeyError:
			pass
	elif utils.check_float(ref) or utils.check_int(ref):
		corrected += ref
	return corrected

def split_and_filter(arr, split_pts, min_length=0):
	"""
	Split continuous array at the specified points.

	Auxilliary function to split continuous current or voltage data into individual holds. Splits the array at the specified indices and discards resulting arrays which are shorter than the required minimum length.

	Parameters
	___________
	arr: list or numpy array
		Array to split
	split_pts: int or array-like
		Indices at which to split the array
	min_length: int (default=0)
		Minimum length of the arrays which result from spliting the intial array. Arrays shorter than this value will be discarded
	
	Returns
	________
	steps: numpy array
		Array containing one array for each hold/step
	"""
	if type(arr) == list:
		arr = np.array(arr)
	steps = np.split(arr, split_pts)
	steps = np.array([s for s in steps if len(s) > min_length])
	return steps

def std_agg(arr):
	"""
	Aggregate standard deviations of multiple measurements.

	Auxilliary function to calculate the aggregate standard deviation of multiple measuremtns. Assumes that the measurements are independent of each other

	Parameters
	___________
	arr: list or numpy array
		Array of standard deviations to be aggregated

	Returns
	________
	sd: float
		Aggregated standard deviation
	"""
	if type(arr) == list:
		arr = np.array(arr)
	sd = np.sqrt(np.sum(arr**2))
	return sd

def std_last_pts(arr, numpts=300):
	"""
	Standard deviation of the last several values of an array

	Auxilliary function to compute the standard deviation of an array accross the last several data points. Useful for obtaining a steady-state standard deviation.

	Parameters
	___________
	arr: list or numpy array
		Array of values used to compute the standard deviation.
	numpts: int (default=300)
		Standard deviation is calculated using the last numpts values of the array.

	Returns
	________
	sd: float
		Standard deviation of the last several values
	"""
	if type(arr) == list:
		arr = np.array(arr)
	sd = np.std(arr[-numpts:], ddof=1)
	return sd